print("import: 'numba'")
import numba

print("import: 'numba.cext'")
import numba.cext

print("import: 'numba.core'")
import numba.core

print("import: 'numba.core.annotations'")
import numba.core.annotations

print("import: 'numba.core.datamodel'")
import numba.core.datamodel

print("import: 'numba.core.rewrites'")
import numba.core.rewrites

print("import: 'numba.core.runtime'")
import numba.core.runtime

print("import: 'numba.core.typeconv'")
import numba.core.typeconv

print("import: 'numba.core.types'")
import numba.core.types

print("import: 'numba.core.typing'")
import numba.core.typing

print("import: 'numba.core.unsafe'")
import numba.core.unsafe

print("import: 'numba.cpython'")
import numba.cpython

print("import: 'numba.cpython.unsafe'")
import numba.cpython.unsafe

print("import: 'numba.cuda'")
import numba.cuda

print("import: 'numba.cuda.cudadrv'")
import numba.cuda.cudadrv

print("import: 'numba.cuda.kernels'")
import numba.cuda.kernels

print("import: 'numba.cuda.simulator'")
import numba.cuda.simulator

print("import: 'numba.cuda.simulator.cudadrv'")
import numba.cuda.simulator.cudadrv

print("import: 'numba.cuda.tests'")
import numba.cuda.tests

print("import: 'numba.cuda.tests.cudadrv'")
import numba.cuda.tests.cudadrv

print("import: 'numba.cuda.tests.cudadrv.data'")
import numba.cuda.tests.cudadrv.data

print("import: 'numba.cuda.tests.cudapy'")
import numba.cuda.tests.cudapy

print("import: 'numba.cuda.tests.cudasim'")
import numba.cuda.tests.cudasim

print("import: 'numba.cuda.tests.nocuda'")
import numba.cuda.tests.nocuda

print("import: 'numba.core.datamodel'")
import numba.core.datamodel

print("import: 'numba.experimental'")
import numba.experimental

print("import: 'numba.experimental.jitclass'")
import numba.experimental.jitclass

print("import: 'numba.misc'")
import numba.misc

print("import: 'numba.misc.help'")
import numba.misc.help

print("import: 'numba.np'")
import numba.np

print("import: 'numba.np.ufunc'")
import numba.np.ufunc

print("import: 'numba.np.unsafe'")
import numba.np.unsafe

print("import: 'numba.parfors'")
import numba.parfors

print("import: 'numba.pycc'")
import numba.pycc

print("import: 'numba.core.rewrites'")
import numba.core.rewrites

print("import: 'numba.roc'")
import numba.roc

print("import: 'numba.roc.hlc'")
import numba.roc.hlc

print("import: 'numba.roc.hsadrv'")
import numba.roc.hsadrv

print("import: 'numba.roc.servicelib'")
import numba.roc.servicelib

print("import: 'numba.roc.tests'")
import numba.roc.tests

print("import: 'numba.roc.tests.hsadrv'")
import numba.roc.tests.hsadrv

print("import: 'numba.roc.tests.hsapy'")
import numba.roc.tests.hsapy

print("import: 'numba.core.runtime'")
import numba.core.runtime

print("import: 'numba.scripts'")
import numba.scripts

print("import: 'numba.stencils'")
import numba.stencils

print("import: 'numba.testing'")
import numba.testing

print("import: 'numba.tests'")
import numba.tests

print("import: 'numba.tests.npyufunc'")
import numba.tests.npyufunc

print("import: 'numba.tests.pycc_distutils_usecase'")
import numba.tests.pycc_distutils_usecase

print("import: 'numba.tests.pycc_distutils_usecase.nested'")
import numba.tests.pycc_distutils_usecase.nested

print("import: 'numba.typed'")
import numba.typed

print("import: 'numba.types'")
import numba.types

print("import: 'numba.core.typing'")
import numba.core.typing

print("import: 'numba.core.unsafe'")
import numba.core.unsafe

